/* Definitions of IRC message functions and list of messages.
 *
 * Auspice Services is copyright (c) 2000-2001 In Mean
 *     E-mail: auspice@auspice.org
 * SirvNET Services is copyright (c) 1998-2000 Trevor Klingbeil.
 *     E-mail: <priority1@dal.net>
 * Originally based on EsperNet Services(c) by Andy Church.
 * This program is free but copyrighted software; see the file LICENSE for
 * details.
 */

#include "../inc/services.h"

// char *server_list[256];
int server_size = 0;
int serv_com = 0;
int day_serv_com = 0;
int hour_serv_com = 0;
int min_serv_com = 0;

static struct server {
   char *name;
   char *prot;
   char *desc;
   char *uplink;
   long clients, maxclients;
} *servers = NULL;

/* List of messages is at the bottom of the file. */
static void m_whois(char *source, int ac, char **av);

/*************************************************************************/
static void m_nickcoll(char *source, int ac, char **av)
{
    do_quit(source, ac, av);
}
/*************************************************************************/
static void m_user(char *source, int ac, char **av)
{
    send_cmd (server_name, "446 %s :USERS has been disabled", source);
}
/*************************************************************************/
static void m_umode2(char *source, int ac, char **av)
{
        av[1] = sstrdup (av[0]);
        strcpy (av[0], source);
        do_umode (source, ac, av);
}
/*************************************************************************/
static void m_netinfo(char *source, int ac, char **av)
{
#ifdef UNREAL
    send_cmd (server_name, "NETINFO 0 %lu 2302 * * * * :%s",
                time(NULL), av[7]);
#endif
}
/*************************************************************************/
static void m_time(char *source, int ac, char **av)
{
        time_t t;
        struct tm *tm1;
        char buf[94];

        time(&t);
        tm1 = localtime(&t);
        strftime(buf, sizeof(buf), "%A %B %d %Y -- %H:%M %Z", tm1);

        send_cmd (server_name, "391 %s %s :%s", source, server_name, buf);

}
/*************************************************************************/
static void m_stats(char *source, int ac, char **av)
{
    time_t uptime = time(NULL) - start_time;
    long count, count2, mem, mem2;
    int tmpnow;
    tmpnow = time(NULL) - start_time;

    if (ac != 2)
        return;

     switch(*av[0]) {

      case 'u':
      case 'U':
        send_cmd (server_name, "242 %s :Services up %d day%s, %d:%02d:%02d",
                   source, tmpnow/86400, (tmpnow/86400 == 1) ? "" : "s",
                   (tmpnow/3600)%24, (tmpnow/60)%60, tmpnow%60);
        send_cmd(NULL, "250 %s :Current users: %d (%d ops); maximum %d",
                source, usercnt, opcnt, maxusercnt);
	send_cmd(NULL, "NOTICE %s :Connects in total: %li", source, connectscnt);
	send_cmd (server_name, "219 %s %s :End of /STATS report", source, av[0]);
        break;

    case 'S':
    case 's':
          send_cmd(server_name, "372 %s :%s", source, server_name);
          break;
    case 'M':
    case 'm':
          if (!is_oper(source))
             return;
          send_cmd(server_name, "372 %s :%d minute", source, min_serv_com);
          send_cmd(server_name, "372 %s :%d hour", source, hour_serv_com);
          send_cmd(server_name, "372 %s :%d day", source, day_serv_com);
          send_cmd(server_name, "372 %s :%d commands", source, serv_com);
          break;

     case 'R':
     case 'r':

         if (!is_oper(source))
             return;

         get_nickserv_stats(&count, &mem);
         get_chanserv_stats(&count2, &mem2);
         send_cmd(server_name, "372 %s :%d/%d nicks/chans",
              source, count, count2);
         send_cmd(server_name, "372 %s :%d/%d kb",
              source, mem, mem2);

         break;

     } 

}
/*************************************************************************/
void do_infoservices(const char *source, char *nick)
{
     notice(s_NickServ, source,
               "(\2Currently on IRC\2) - \"/whois %s\" for more information", nick);
     notice(s_NickServ, source,
                "\2%s\2 is \"Services Bot\"\n", nick);
}
/*************************************************************************/
static void m_setname(char *source, int ac, char **av)
{
        User *u;
        NickInfo *ni, *hni;

        /* First, update the user structure. */
        u = finduser (source);
	if (!u)
		return;
	if (u) 
	        u->realname = sstrdup(av[0]);

        /* Now, update the nickinfo. */
        ni = findnick (u->nick);
        hni = host(ni);
        if ((ni && ni->flags & NI_IDENTIFIED) || (hni && hni->flags & NI_IDENTIFIED)) {
		free(hni->last_realname);
	        hni->last_realname = smalloc(strlen(u->realname)+1);
		sprintf (hni->last_realname,"%s", u->realname);
	}
}
/*************************************************************************/
static void m_userhost(char *source, int ac, char **av)
{
        User *u;
        NickInfo *ni, *hni;
	char *nick = strtok(av[1],"=");
	char *virthost;

	if (!nick)
		return;

	if (strchr(nick, '*')) {
		virthost = strtok(NULL,"@");
		virthost = strtok(NULL," ");
		nick = strtok(nick, "*");
	} else {
		virthost = strtok(NULL,"@");
		virthost = strtok(NULL," ");
	}

	if (!virthost)
		return;

        u = finduser (nick);

	if (!u)
		return;

	if (u && virthost) 
	        u->host = sstrdup(virthost);
}
/*************************************************************************/
static void m_sethost(char *source, int ac, char **av)
{
        User *u;
        NickInfo *ni, *hni;

        /* First, update the user structure. */
        u = finduser (source);
	if (!u)
		return;
	if (u)
	        u->host = sstrdup(av[0]);

        /* Now, update the nickinfo. */
        ni = findnick (u->nick);
	hni = host(ni);
        if ((ni && ni->flags & NI_IDENTIFIED) || (hni && hni->flags & NI_IDENTIFIED)) {
	    free(hni->last_usermask);
            hni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
 	    sprintf (hni->last_usermask,"%s@%s", u->username, u->host);
	}
}
/*************************************************************************/
static void m_setident(char *source, int ac, char **av)
{
        User *u;
        NickInfo *ni, *hni;

        /* First, update the user structure. */
        u = finduser (source);
	if (!u)
		return;
	if (u)
	        u->username = sstrdup(av[0]);

        /* Now, update the nickinfo. */
        ni = findnick (u->nick);
	hni = host(ni);
        if ((ni && ni->flags & NI_IDENTIFIED) || (hni && hni->flags & NI_IDENTIFIED)) {
	    free(hni->last_usermask);
            hni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
 	    sprintf (hni->last_usermask,"%s@%s", u->username, u->host);
	}
}
/*************************************************************************/
static void m_server(char *source, int ac, char **av)
{
    int i;
    if (ac < 1)
        return;

    if (numserver >= server_size) {
        if (server_size < 8)
            server_size = 8;
        else
            server_size *= 2;

        servers = srealloc(servers, sizeof(*servers) * server_size);
    }
    if (numserver > 0) {
       for (i = 0; i < numserver; i++) {
          if (stricmp(servers[i].name, av[0]) == 0)
            return;
       }
    }

    servers[numserver].name = sstrdup(av[0]);
    servers[numserver].desc = sstrdup(av[2]);
    servers[numserver].uplink = sstrdup(source);

    numserver++;
    check_jupes();
    check_zlines();
}

/*************************************************************************/
void do_stats_server(const char *source)
{
     int i;
     notice(s_StatServ, source, "\2Server Listing\2");
     for (i=0; i<numserver; i++) {
             notice(s_StatServ, source, "%d) %s", i+1, servers[i].name);
     }
     notice(s_StatServ, source, "\2End of Server Listing\2");
}
/*************************************************************************/
static void m_squit(char *source, int ac, char **av)
{
    int i, x=0;
    if (ac < 1)
	return;
    for(i=0;i<=numserver-1;i++)
	if(stricmp(servers[i].name,av[0])==0) {
		free(servers[i].name);
		free(servers[i].desc);
		free(servers[i].uplink);
		for(;i<=numserver-1;i++)
			servers[i]=servers[i+1];
		free(servers[numserver].name);
		free(servers[numserver].desc);
		free(servers[numserver].uplink);
	check_jupe_squit(source, av[0]);
	numserver--;
    }
}

/*************************************************************************/
static void m_ping(char *source, int ac, char **av)
{
    if (ac < 1)
	return;
    send_cmd(server_name, "PONG %s %s", ac>1 ? av[1] : server_name, av[0]);
}

/*************************************************************************/
static void m_away(char *source, int ac, char **av)
{
    if (ac == 0 || *av[0] == 0)		/* un-away */
	check_memos(source);
}

/*************************************************************************/
static void m_join(char *source, int ac, char **av)
{
    if (ac != 1)
	return;
    do_join(source, ac, av);
}

/*************************************************************************/

static void m_sjoin(char *source, int ac, char **av)
{
   do_sjoin(source, ac, av);
}

/*************************************************************************/
void do_statss(const char *source)
{
     int i;
     notice(s_StatServ, source, "There are \2%d\2 servers", numserver);
     for (i=0; i<numserver; i++) {
	     notice(s_StatServ, source, "+ %s", servers[i].name);
	     notice(s_StatServ, source, "-- Desc     : %s", servers[i].desc);
	     notice(s_StatServ, source, "-- Link with: %s", servers[i].uplink);
     }
}
/*************************************************************************/
static void m_kick(char *source, int ac, char **av)
{
    if (ac != 3)
	return;
    if (stricmp(av[1],s_ChanServ)==0) {
	send_cmd(s_OperServ, "JOIN %s", av[0]);
	send_cmd(s_OperServ, "MODE %s +o %s", av[0], av[1]);
    }
    do_kick(source, ac, av);
}

/*************************************************************************/
static void m_kill(char *source, int ac, char **av)
{
    if (ac != 2)
	return;
    do_kill(source, ac, av);
}

/*************************************************************************/
static void m_mode(char *source, int ac, char **av)
{
    char *modestr = av[1];
    char *s, *nick;
    int add = 0, i, x=0, x2=0;
    User *u, *u2;

    if (*av[0] == '#' || *av[0] == '&') {

        char newmodes[128], *rlist[13], *rrlist[13];
        char c, *mchange = newmodes;
        char **bv = av, *nv[8], *ov[8], *reop[8];
        int bc = ac, start = 2, ostart = 2, rstart = 2;
        int vmode = 0;

	if (!findchan(av[0])) {
		log("Mode for nonexist channel.");
		return;
	}

        for (i=2; i < 8; i++) {
           ov[i] = NULL;
           reop[i] = NULL;
        }

	if (ac < 2)
	    return;

        s = modestr;
        bc -= 2;
        bv += 2;

        while (*s) {
            int repeat = 0;
            char *mv[3];
            switch(c = *s++) {
               case '+':
                  if (*mchange-1 != '+')
                     *mchange++ = '+';
                  add = 1;
                  break;
               case '-':
                  if (*mchange-1 != '-')
                     *mchange++ = '-';
                  add = 0;
                  break;

               case 'b':
                 mv[0] = av[0];
                 if (add)
                    mv[1] = "+b";
                 else
                    mv[1] = "-b";

                 if (--bc < 0) {
                   log("channel: MODE %s %s: missing parameter for %cb",
                                     av[0], modestr, add ? '+' : '-');
                   break;
                 }
                 mv[2] = *bv++;

                 do_cmode(source, 3, mv);
                 break;

               case 'L':
                 mv[0] = av[0];
                 if (add)
                    mv[1] = "+L";
                 else
                    mv[1] = "-L";

                 if (--bc < 0) {
                   log("channel: MODE %s %s: missing parameter for %cL",
                                     av[0], modestr, add ? '+' : '-');
                   break;
                 }
                 mv[2] = *bv++;

                 do_cmode(source, 3, mv);
                 break;

               case 'f':
                 mv[0] = av[0];
                 if (add)
                    mv[1] = "+f";
                 else
                    mv[1] = "-f";

                 if (--bc < 0) {
                   log("channel: MODE %s %s: missing parameter for %cf",
                                     av[0], modestr, add ? '+' : '-');
                   break;
                 }
                 mv[2] = *bv++;

                 do_cmode(source, 3, mv);
                 break;

               case 'v':
                 mv[0] = av[0];
                 if (add)
                    mv[1] = "+v";
                 else
                    mv[1] = "-v";

                 if (--bc < 0) {
                   log("channel: MODE %s %s: missing parameter for %cv",
                                     av[0], modestr, add ? '+' : '-');
                   break;
                 }
                 mv[2] = *bv++;
                 do_cmode(source, 3, mv);
                 break;

               case 'q':
                 mv[0] = av[0];
                 if (add)
                    mv[1] = "+q";
                 else
                    mv[1] = "-q";

                 if (--bc < 0) {
                   log("channel: MODE %s %s: missing parameter for %cq",
                                     av[0], modestr, add ? '+' : '-');
                   break;
                 }
                 mv[2] = *bv++;
                 do_cmode(source, 3, mv);
                 break;

               case 'a':
                 mv[0] = av[0];
                 if (add)
                    mv[1] = "+a";
                 else
                    mv[1] = "-a";

                 if (--bc < 0) {
                   log("channel: MODE %s %s: missing parameter for %ca",
                                     av[0], modestr, add ? '+' : '-');
                   break;
                 }
                 mv[2] = *bv++;
                 do_cmode(source, 3, mv);
                 break;


               case 'o':
                 mv[0] = av[0];
                 mv[1] = "+o";
                 repeat = 0;

                 if (--bc < 0) {
                   log("channel: MODE %s %s: missing parameter for %co",
                                     av[0], modestr, add ? '+' : '-');
                   break;
                 }

                 nick = *bv++;

                 for (i=0; i < x; i++) {
                    if (rlist[i] && stricmp(rlist[i], nick) == 0)
       	                repeat++;
                 }

                 rlist[x++] = nick;

                 mv[2] = nick;

                 if (add) {
                     u = finduser(nick);

                     if (!u)
                        break;

                     if (!check_valid_op(u, av[0], 0) &&
                        stricmp(source, u->nick) && !repeat) {
                        ov[ostart++] = nick;
	                mv[1] = "-o";
                        do_cmode(source, 3, mv);
                        break;
                     } else {
                       if (!repeat)
                          do_cmode(source, 3, mv);
		     }

                 } else {
                     ChannelInfo *ci;

                     mv[1] = "-o";
                     ci = cs_findchan(av[0]);
                     u = finduser(nick);
	             u2 = finduser(source);

                     if (!u || !u2)
                        break;

                     for (i=0; i < x2; i++) {
                        if (rrlist[i] && stricmp(rrlist[i], nick) == 0)
                           repeat++;
                     }

                     rrlist[x2++] = nick;

                     if (ci && (get_access(u, ci) > 4) && !repeat) {
				if (u2 && (get_access(u2, ci) < 5))
		                        reop[rstart++] = nick;
				else
		                        do_cmode(source, 3, mv);
                     } else {
                         do_cmode(source, 3, mv);
		     }
                 }
                 break;

              default:

                   *mchange++ = c;
                   if (c == 'l' && add)
                         nv[start++] = *bv++;
                   else if (c == 'k')
                      nv[start++] = *bv++;
                   else vmode = 1;
                   break;
            }
       }

       if (start > 2 || vmode) {
           nv[0] = av[0];
           nv[1] = newmodes;
           do_cmode(source, start-1, nv);
       }

       if (ov[2])
          send_cmd(s_ChanServ, "MODE %s -%s%s%s%s%s%s %s %s %s %s %s %s",
                 av[0], "o", ov[3] ? "o" : "",
                 ov[4] ? "o" : "", ov[5] ? "o" : "",
                 ov[6] ? "o" : "", ov[7] ? "o" : "",
                 ov[2], ov[3] ? ov[3] : "",
                 ov[4] ? ov[4] : "", ov[5] ? ov[5] : "",
                 ov[6] ? ov[6] : "", ov[7] ? ov[7] : "");

       if (reop[2])
          send_cmd(s_ChanServ, "MODE %s +%s%s%s%s%s%s %s %s %s %s %s %s",
                 av[0], "o", reop[3] ? "o" : "",
                 reop[4] ? "o" : "", reop[5] ? "o" : "",
                 reop[6] ? "o" : "", reop[7] ? "o" : "",
                 reop[2], reop[3] ? reop[3] : "",
                 reop[4] ? reop[4] : "", reop[5] ? reop[5] : "",
                 reop[6] ? reop[6] : "", reop[7] ? reop[7] : "");
    } else {
	if (ac != 2)
	    return;
	do_umode(source, ac, av);
    }
}

/*************************************************************************/
static void m_motd(char *source, int ac, char **av)
{
    FILE *f;
    char buf[BUFSIZE];
    char buf2[BUFSIZE];
    char *x;
    int i, all=0, avail=0;
    time_t uptime = time(NULL) - start_time;
    User *u = finduser(source);

    if (u && (stricmp(tokn[19], "ON") ==0))
        do_update_flood(u, 2);

    f = fopen(MOTD_FILENAME, "r");
    send_cmd(server_name, "375 %s :- %s Message of the Day",
		source, server_name);
   if (f) {
	while (fgets(buf, sizeof(buf), f)) {
	    buf[strlen(buf)-1] = 0;
	    send_cmd(server_name, "372 %s :- %s", source, buf);
	}
	fclose(f);
    }

    send_cmd(server_name, "372 %s :-", source);

    send_cmd(server_name, "372 %s :- This list of people can retreive"
                " lost passwords:", source);

    send_cmd(server_name, "372 %s :-", source);

    *buf = 0;
    *buf2 = 0;

    for (i = 0; i < MAX_SERVADMINS; i++) {
        if (services_admins[i]) {
            all++;
            if (finduser(services_admins[i])) {
                strcat(buf2, services_admins[i]);
                strcat(buf2, " ");
                avail++;
            }
            if (!finduser(services_admins[i])) {
                strcat(buf, services_admins[i]);
                strcat(buf, " ");
            }
        }
    }
    for (i = 0; i < MAX_SERVROOTS; i++) {
        if (services_roots[i]) {
            all++;
            if (finduser(services_roots[i])) {
                strcat(buf2, services_roots[i]);
                strcat(buf2, " ");
                avail++;
            }
            if (!finduser(services_roots[i])) {
                strcat(buf, services_roots[i]);
                strcat(buf, " ");
            }
        }
    }
    if (!*buf && !*buf2)
        strcpy(buf, "List: Not Currently Available");

    send_cmd(server_name, "372 %s :- \2%s \n\2", source, buf2);
    send_cmd(server_name, "372 %s :-", source);
    send_cmd(server_name, "372 %s :- %s \n", source, buf);

    send_cmd(server_name, "372 %s :-", source);
    send_cmd(server_name, "372 %s :- %d of %d nicks available.",
         source, avail, all);

    send_cmd(server_name, "372 %s :- Note: The nicks that are in \2bold\2"
                 " are currently online", source);

    send_cmd(server_name, "372 %s :-", source);

    if (MOTD_UPDATE && !readonly) {
       i = (UPDATE_TIMEOUT - (time(NULL) - last_update));
       send_cmd(server_name, "372 %s :- Next DataBase Update: %d Minute%s, %d second%s",
           source, i/60, i==1 ? "" : "s",
           i%60, i==1 ? "" : "s");
       i=0;
    }

    if (MOTD_UPTIME) {
      if (uptime > 86400)
          send_cmd(server_name, "372 %s :- "
                        "Services up \2%d\2 day%s, \2%02d:%02d\2", source,
                        uptime/86400, (uptime/86400 == 1) ? "" : "s",
                        (uptime/3600) % 24, (uptime/60) % 60);
      else if (uptime > 3600)
          send_cmd(server_name, "372 %s :- "
                        "Services up \2%d hour%s, %d minute%s\2", source,
                        uptime/3600, uptime/3600==1 ? "" : "s",
                        (uptime/60) % 60, (uptime/60)%60==1 ? "" : "s");
      else
          send_cmd(server_name, "372 %s :- "
                        "Services up \2%d minute%s, %d second%s\2", source,
                        uptime/60, uptime/60==1 ? "" : "s",
                        uptime%60, uptime%60==1 ? "" : "s");
    }


    send_cmd(server_name, "372 %s :-", source);

    send_cmd(server_name, "372 %s :- Auspice Services are copyright (c) "
                "2000-2001 In Mean.", source);
    send_cmd(server_name, "376 %s :End of /MOTD command.", source);
}
/*************************************************************************/
static void m_nick(char *source, int ac, char **av)
{
#if !defined(BAHAMUT) && !defined(NICKV2)
    if (ac == 8) {
	/* Get rid of the useless extra parameter. */
	av[6] = av[7];
	ac--;
    }

    if ((!*source && ac != 7) || (*source && ac != 2))
	return;
#endif
    do_nick(source, ac, av);
}
/*************************************************************************/
static void m_part(char *source, int ac, char **av)
{
    if (ac < 1 || ac > 2)
	return;
    do_part(source, ac, av);
}

/*************************************************************************/
static void m_privmsg(char *source, int ac, char **av)
{
    time_t starttime, stoptime;	/* When processing started and finished */
    time_t now = time(NULL);
    char *nick;
    char *tnick;
    User *u;
    u = finduser(source);

    if (ac != 2)
	return;

    nick = sstrdup(av[0]);
    if(strchr(nick, '@')) {
	tnick = strtok(nick, "@");
	av[0] = sstrdup(tnick);
    }

    if (!(*av[0] == '#') && (stricmp(tokn[19], "ON") == 0))
        do_update_flood(u, 0);


    /* Check if we should ignore.  Operators always get through. */
    if ((allow_ignore == 1) && !is_oper(source) && !(*av[0] == '#')) {
         if (check_ignore(u->nick, u->username, u->host) != 0)
              return;
    }

    if (u && (u->flags & U_NOSERV)) {
	log("Access denied to services from %s", source);
	return;
    }

    if ((debug == 1) && ((stricmp(av[0], s_OperServ) == 0)
       || (stricmp(av[0], s_OperServ) == 0)
       || (stricmp(av[0], s_RootServ) == 0)
       || (stricmp(av[0], s_NickServ) == 0)
       || (stricmp(av[0], s_MemoServ) == 0)
       || (stricmp(av[0], s_ChanServ) == 0)
       || (stricmp(av[0], s_HelpServ) == 0)
       || (stricmp(av[0], s_Agent) == 0)
       || (stricmp(av[0], s_StatServ) == 0)
       || (stricmp(av[0], s_ServicesX) == 0)
       || (stricmp(av[0], s_ServicesW) == 0)
       || (stricmp(av[0], s_SpamChecker) == 0)
#ifdef DEFHOSTSERV
       || (stricmp(av[0], s_HostServ) == 0)
#endif
       || (stricmp(av[0], s_MassServ) == 0)
       || (stricmp(av[0], s_BotServ) == 0))
       && (!stricmp(av[1], "*identify*") ==0))
            log("%s: %s", source, inbuf);

    if ((sndebug == 1) && ((stricmp(av[0], s_OperServ) == 0)
       || (stricmp(av[0], s_OperServ) == 0)
       || (stricmp(av[0], s_RootServ) == 0)
       || (stricmp(av[0], s_NickServ) == 0)
       || (stricmp(av[0], s_MemoServ) == 0)
       || (stricmp(av[0], s_ChanServ) == 0)
       || (stricmp(av[0], s_HelpServ) == 0)
       || (stricmp(av[0], s_Agent) == 0)
       || (stricmp(av[0], s_StatServ) == 0)
       || (stricmp(av[0], s_ServicesX) == 0)
       || (stricmp(av[0], s_ServicesW) == 0)
#ifdef DEFHOSTSERV
       || (stricmp(av[0], s_HostServ) == 0)
#endif
       || (stricmp(av[0], s_MassServ) == 0)
       || (stricmp(av[0], s_BotServ) == 0)
       || (stricmp(av[0], s_SpamChecker) == 0))
       && (!stricmp(av[1], "*identify*") ==0))
           send_cmd(s_OperServ, "PRIVMSG %s :%s: %s", snoopchan, source, inbuf);

    if (!(*av[0] == '#')) {

    serv_com += 1;

    if (now > d_check_time) {
       d_check_time = time(NULL) + 86400;
       day_serv_com = 1;
    } else {
        day_serv_com += 1;
    }

    if (now > h_check_time) {
       h_check_time = time(NULL) + 3600;
       hour_serv_com = 1;
    } else {
        hour_serv_com += 1;
    }
   if (now > m_check_time) {
      m_check_time = time(NULL) + 60;
      min_serv_com = 1;
   } else {
       min_serv_com += 1;
   }

    starttime = time(NULL);
   }

    if (stricmp(av[0], s_OperServ) == 0) {
        if (is_oper(source) && !(u->flags & U_ABUSIVE))
	    operserv(source, av[1]);
    } else if (stricmp(av[0], s_RootServ) == 0) {
        if (is_services_root(source) && is_oper(source))
            rootserv(source, av[1]);
	else {
               notice(s_RootServ, source, "Access denied.");
	}
    } else if (stricmp(av[0], s_NickServ) == 0) {
	nickserv(source, av[1]);
    } else if (stricmp(av[0], s_ChanServ) == 0) {
	chanserv(source, av[1]);
    } else if (stricmp(av[0], s_MemoServ) == 0) {
	memoserv(source, av[1]);
    } else if (stricmp(av[0], s_Agent) == 0) {
	agent(source, av[1]);
    } else if (stricmp(av[0], s_HelpServ) == 0) {
	helpserv(s_HelpServ, source, av[1]);
    } else if (stricmp(av[0], s_StatServ) == 0) {
	statserv(source, av[1]);
    } else if (stricmp(av[0], s_ServicesX) == 0) {
	servicesX(source, av[1]);
    } else if (stricmp(av[0], s_ServicesW) == 0) {
	servicesW(source, av[1]);
    } else if (stricmp(av[0], s_SpamChecker) == 0) {
	SpamCheck(source, av[1]);
#ifdef DEFHOSTSERV
    } else if (stricmp(av[0], s_HostServ) == 0) {
	HostServ(source, av[1]);
#endif
    } else if (stricmp(av[0], s_MassServ) == 0) {
	massserv(source, av[1]);
    } else if (stricmp(av[0], s_BotServ) == 0) {
	sbotserv(source, av[1]);
    } else if (check_ifbot(av[0])) {
        botserv(source, av[0], av[1]);
    } else if (*av[0] == '#') {
    	if (*av[1] == '~') {
		chanserv(source, av[1]);
	} else {
		bmsgchan(source, av[0], av[1]);
	}
    } else {
	if (findchan(snoopchan))
		send_cmd(s_OperServ, "PRIVMSG %s :<%s/%s> %s", snoopchan,  source, av[0], av[1]);
    }
    /* Add to ignore list if the command took a significant amount of time. */
/*
    if (allow_ignore) {
	stoptime = time(NULL);
	if (stoptime > starttime && *source && !strchr(source, '.'))
            add_ignore(mask, source, 10);
    }
*/
}

/*************************************************************************/

static void m_quit(char *source, int ac, char **av)
{
    if (ac != 1)
	return;
    do_quit(source, ac, av);
}

/*************************************************************************/

static void m_topic(char *source, int ac, char **av)
{
    if (ac != 4)
	return;
    do_topic(source, ac, av);
}

/*************************************************************************/

Message messages[] = {
    { "436",       m_nickcoll },
#ifdef LIQUID
    { "449",       NULL },
    { "CHOST",     NULL },
    { "CIDENT",     NULL },
    { "SERVERINFO", NULL },
    { "IJOIN", m_join },
#endif
    { "380",	   NULL },
    { "401",	   NULL },
    { "402",	   NULL },
    { "302",	   m_userhost },
    { "HELP",	   NULL },
    { "4",	   NULL },
    { "AWAY",      m_away },
    { "6",      m_away },
    { "JOIN",      m_join },
    { "C",      m_join },
#ifdef SJOIN
    { "SJOIN",     m_sjoin },
    { "~",     m_sjoin },
#endif
    { "KICK",      m_kick },
    { "H",      m_kick },
    { "KILL",      m_kill },
    { ".",      m_kill },
    { "SVSKILL",   m_kill },
    { "h",   m_kill },
    { "MODE",      m_mode },
    { "G",      m_mode },
    { "SVSMODE",   m_mode },
    { "n",   m_mode },
    { "MOTD",      m_motd },
    { "F",      m_motd },
    { "NICK",      m_nick },
    { "&",      m_nick },
    { "SVSNICK",   m_nick },
    { "e",   m_nick },
    { "NOTICE",    NULL },
    { "B",    NULL },
    { "PART",      m_part },
    { "D",      m_part },
    { "PASS",      NULL },
    { "<",      NULL },
    { "PING",      m_ping },
    { "8",      m_ping },
    { "PRIVMSG",   m_privmsg },
    { "!",   m_privmsg },
    { "QUIT",      m_quit },
    { ",",      m_quit },
    { "SERVER",    m_server },
    { "'",    m_server },
    { "SQUIT",     m_squit },
    { "-",     m_squit },
    { "TOPIC",     m_topic },
    { ")",     m_topic },
    { "WALLOPS",   NULL },
    { "=",   NULL },
    { "VERSION",   m_version },
    { "+",   m_version },
    { "AKILL",     NULL },
    { "V",     NULL },
    { "GLOBOPS",   NULL },
    { "]",   NULL },
    { "CHATOPS",   NULL },
    { "p",   NULL },
    { "GNOTICE",   NULL },
    { "Z",   NULL },
    { "GOPER",     NULL },
    { "[",     NULL },
    { "RAKILL",    NULL },
    { "Y",    NULL },
    { "ADMIN",     m_admin },
    { "@",     m_admin },
    { "NETINFO",   NULL },
    { "AO",   m_netinfo },
    { "SETNAME",   m_setname },
    { "AE",   m_setname },
    { "USERS",     m_user },
    { "3",     m_user },
    { "INFO",      m_info },
    { "/",      m_info },
    { "CREDITS",   m_credits },
    { "AJ",   m_credits },
    { "STATS",     m_stats },
    { "2",     m_stats },
    { "TIME",      m_time },
    { ">",      m_time },
    { "WHOIS",    NULL },
    { "#",     NULL },
    { "SETHOST",   m_sethost },
    { "AA",   m_sethost },
    { "SETIDENT",  NULL },
    { "AD",  NULL },
    { "CHGHOST",   NULL },
    { "AL",   NULL },
    { "SMO",       NULL },
    { "AU",       NULL },
    { "PROTOCTL",  NULL },
    { "_",  NULL },
    { "SQLINE",  NULL },
    { "c",  NULL },
    { "UNSQLINE",  NULL },
    { "d",  NULL },
    { "SVSNOOP",  NULL },
    { "f",  NULL },
    { "TRACE",  NULL },
    { "b",  NULL },
    { "ERROR",  NULL },
    { "SWHOIS",  NULL },
    { "VCTRL",  NULL },
    { "SYNC",  NULL },
    { "SETIDENT",  m_setident },
    { "AD",  m_setident },
#ifdef UNREAL
    { "UMODE2",  m_umode2 },
    { "|",  m_umode2 },
#endif
#ifdef ULTIMATE
    { "NETG",  NULL },
    { "GCLIENT",  NULL },
    { "SNETINFO",   NULL },
#endif
    { NULL }
};

/*************************************************************************/

Message *find_message(const char *name)
{
    Message *m;

    for (m = messages; m->name; m++) {
	if (stricmp(name, m->name) == 0)
	    return m;
    }
    return NULL;
}

/*************************************************************************/
static void m_whois(char *source, int ac, char **av)
{
    char *clientdesc;

    if (source && ac >= 1) {
        if (stricmp(av[0], s_NickServ) == 0)
	    strcpy(clientdesc, "Nickname Services");
        else if (stricmp(av[0], s_ChanServ) == 0)
	    strcpy(clientdesc, "Channel Services");
        else if (stricmp(av[0], s_MemoServ) == 0)
	    strcpy(clientdesc, "Memo Services");
        else if (stricmp(av[0], s_HelpServ) == 0)
	    strcpy(clientdesc, "Help Services");
        else if (stricmp(av[0], s_OperServ) == 0)
	    strcpy(clientdesc, "Operator Services");
        else if (stricmp(av[0], s_RootServ) == 0)
	    strcpy(clientdesc, "Services Root System");
        else if (stricmp(av[0], s_GlobalNoticer) == 0)
	    strcpy(clientdesc, "Global Noticer");
        else if (stricmp(av[0], s_ServicesW) == 0)
	    strcpy(clientdesc, "Nickname Services Agent");
        else if (stricmp(av[0], s_ServicesX) == 0)
	    strcpy(clientdesc, "Channel Services Agent");
        else if (stricmp(av[0], s_Agent) == 0)
	    strcpy(clientdesc, "General Services Agent");
        else {
            send_cmd(server_name, "401 %s %s :No such service.", source, av[0]);
            return;
        }
        send_cmd(server_name, "311 %s %s %s %s :%s", source, av[0],
                server_name, services_host, clientdesc);
        send_cmd(server_name, "312 %s %s %s :%s", source, av[0],
                server_name, server_desc);
        send_cmd(server_name, "318 End of /WHOIS response.");
	}
}
